#include "evaluate.h"


const int       shield[8] = {0, 32, 16, 8, 2, 1, 1, 0};
const int       storm[8] = {6, 1, 2, 4, 8, 16, 16, 16};

/*
	Queen near King		20	4
	Ropen by King		15	3
	Rhopen by King		5	1
	Qopen by King		5	1
	Qhopen by King
	pawn				20	4
	Outpost				15	3
	Outpost protected	35	7
	Marauder			40	8
	Qarauder			30	4
		
	notes:
		* marauder on open and near king is 55!
		* marauder AND knight outpost ought to be strong: +5 +10 ?
		* q on 7th on open and near king 45?
		* must _ASSUME_ weakest/uncoordinated attacks,
		* until multiple attackers are overwhelming
*/



/*
 * 0,   1    2    3    4    5    6    7    8    9   10   11   12   13   14
 * 15
 */
const int       menace[64] = {0, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50, 55, 60, 65, 75, 85,
	95, 105, 115, 125, 135, 145, 155, 165, 175, 185, 195, 205, 215, 225, 235, 245,
	250, 255, 260, 265, 270, 275, 280, 285, 290, 300, 305, 310, 315, 320, 325, 330,
335, 340, 345, 350, 355, 360, 365, 370, 375, 380, 385, 390, 395, 400, 405, 410};
const int       inactive[32] = {0, 10, 20, 31, 45, 60, 75, 100, 125, 150, 175, 200, 225, 250, 275, 300,
325, 350, 375, 400, 425, 450, 475, 500, 525, 550, 575, 600, 625, 650, 675, 700};

const int       kingpawn[8] = {10, 15, 5, 2, 1, 0, 0, 0};

int             best = 0;

int 
evaluate ()
{
	int             sq, i, p, f, r;

	int             lap[2][10] = {{7, 7, 7, 7, 7, 7, 7, 7, 7, 7}, {0, 0, 0, 0, 0, 0, 0, 0, 0, 0}};	/* black default=7 white
													 * default=0 */
	int             lar[2][10] = {{7, 7, 7, 7, 7, 7, 7, 7, 7, 7}, {0, 0, 0, 0, 0, 0, 0, 0, 0, 0}};
	int             map[2][10] = {{0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, {7, 7, 7, 7, 7, 7, 7, 7, 7, 7}};
	int             bishopb[] = {0, 0};
	int             bishopw[] = {0, 0};
	int             obstaclesb = 0;
	int             obstaclesw = 0;
	int             weakb = 0;
	int             weakw = 0;
	int             passerb = 0;
	int             passerw = 0;
	int             inactiveb = 0;
	int             inactivew = 0;
	int             rooksb = 0;
	int             rooksw = 0;
	int             safetyb = 0;
	int             safetyw = 0;
	int             offenseb = 0;
	int             offensew = 0;
	int             menaceb = 0;
	int             menacew = 0;
	int             homer = 0;	/* pawn on home row. black=bit0,
					 * white bit1 */
	int             t = 1;
	int             s;

	sEHASHENTRY    *ehash;

	if ((ehash = eprobe ())) {
		return stm ? ehash->score + 5 : -ehash->score + 5;
	}
	++evals;

	s = material (&t);

	if (t)
		return ((stm) ? s : -s);	/* draw[ish] */

	/* Pass 1 */
	/* BLACK */
	for (i = 0; i < 16; ++i) {
		sq = pieces[BLACK][i];
		if (sq) {
			p = board[sq];
			if (p == BP) {
				f = FILE (sq);
				r = RANK (sq);
				if (r < lap[BLACK][f + 1])
					lap[BLACK][f + 1] = r;
				if (r > map[BLACK][f + 1])
					map[BLACK][f + 1] = r;
				homer |= (r == 1);
			} else if (p == BB)
				++bishopb[sq_color[sq]];
			else if (p == BR) {
				f = FILE (sq);
				r = RANK (sq);
				if (r < lar[BLACK][f + 1])
					lar[BLACK][f + 1] = r;
			}
		}
	}
	/* WHITE */
	for (i = 0; i < 16; ++i) {
		sq = pieces[WHITE][i];
		if (sq) {
			p = board[sq];
			if (p == WP) {
				f = FILE (sq);
				r = RANK (sq);
				if (r > lap[WHITE][f + 1])
					lap[WHITE][f + 1] = r;
				if (r < map[WHITE][f + 1])
					map[WHITE][f + 1] = r;
				homer |= ((r == 6) << 1);
			} else if (p == WB)
				++bishopw[sq_color[sq]];
			else if (p == WR) {
				f = FILE (sq);
				r = RANK (sq);
				if (r > lar[WHITE][f + 1])
					lar[WHITE][f + 1] = r;
			}
		}
	}

	/*
	printf("           A B C D E F G H\n");
	printf("lap[BLACK] ");
	for(i=0 ; i<8; ++i) printf("%d ", lap[BLACK][i+1]);
	printf("\n");
	printf("map[BLACK] ");
	for(i=0 ; i<8; ++i) printf("%d ", map[BLACK][i+1]);
	printf("\n");
	printf("lap[WHITE] ");
	for(i=0 ; i<8; ++i) printf("%d ", lap[WHITE][i+1]);
	printf("\n");
	printf("map[WHITE] ");
	for(i=0 ; i<8; ++i) printf("%d ", map[WHITE][i+1]);
	printf("\n");
	*/

	/* Pass 2 */
	/* BLACK */
	for (i = 0; i < 16; ++i) {
		sq = pieces[BLACK][i];
		if (sq) {
			p = board[sq];

			switch (p) {
			case BP:
				f = FILE (sq);
				r = RANK (sq);
				t = (r != map[BLACK][f + 1]);	/* doubled */
				weakb += 8 * t;
				if (t == 0 && r >= lap[WHITE][f + 1 - 1] && r >= lap[WHITE][f + 1 + 1] && r > lap[WHITE][f + 1]) {
					passerb += 10 * (passerb > 0);	/* multiple passer bonus */
					passerb += passed_pawn[r];	/* base value */
					passerb += connected_passer[r] * (board[sq - 1] == BP || board[sq + 1] == BP || board[sq - 17] == BP || board[sq - 15] == BP);
					passerb -= blocked_passer[r] * (IS_WHITE (board[sq + 16]));	/* malis: blocked */
					passerb += (15 + 4 * r) * (lar[BLACK][f + 1] < r);	/* bonus: rook behind */
					passerb += escorted_passer[r] * (distance[128 + sq + 16 - kingsq[BLACK]] == 1);	/* bonus: king escort */
					passerb -= escorted_passer[r] * (distance[128 + sq + 16 - kingsq[WHITE]] == 1);	/* malis: nme king */
				}
				weakb += 15 * (r < lap[BLACK][f + 1 - 1] && r < lap[BLACK][f + 1 + 1]);
				inactiveb += (f > 1 && f < 6 && r == 1 && board[sq + 16] != MT);
				menaceb += 4 * (f > 1 && f < 6 && r > 2);
				obstaclesb += bishopb[sq_color[sq]] * middle[sq];
				offenseb += kingpawn[distance[128 + sq + 16 - kingsq[BLACK]]];
				offenseb -= kingpawn[distance[128 + sq + 16 - kingsq[WHITE]]];
				/* spaceb += 3 * r; */
				break;
			case BN:
				f = FILE (sq);
				r = RANK (sq);
				if (((f > 1 && f < 6 && r > 2 && r < 7) || distance[128 + sq - kingsq[WHITE]] < 4) && lap[WHITE][f + 1 - 1] <= r && lap[WHITE][f + 1 + 1] <= r) {
					menaceb += 2;
					if (board[sq - 17] == BP || board[sq - 15] == BP) {
						menaceb += 3;
					}
				}
				inactiveb += (f < 1) + (f > 6) + 2 * (r < 1) + (r > 6);
				break;
			case BB:
				f = FILE (sq);
				r = RANK (sq);
				if (((f > 1 && f < 6 && r > 2 && r < 7) || distance[128 + sq - kingsq[WHITE]] < 4) && lap[WHITE][f + 1 - 1] <= r && lap[WHITE][f + 1 + 1] <= r) {
					menaceb += 2;
					if (board[sq - 17] == BP || board[sq - 15] == BP) {
						menaceb += 3;
					}
				}
				inactiveb += (f < 1) + (f > 6) + 2 * (r < 1) + (r > 6);
				break;
			case BR:
				f = FILE (sq);
				r = RANK (sq);
				if (lap[BLACK][f + 1] > 5) {	/* lap: black default=7
								 * white default=0 */
					if (r >= lap[WHITE][f + 1]) {
						rooksb += 15;
						menaceb += (abs (f - FILE (kingsq[WHITE])) < 3) * 3;
					} else
						menaceb += (abs (f - FILE (kingsq[WHITE])) < 3);
				} else if (r >= map[BLACK][f + 1]) {	/* map: black default=0
									 * white default=7  */
					if (r >= lap[WHITE][f + 1]) {
						/* rooksb += 10; */
						menaceb += (abs (f - FILE (kingsq[WHITE])) < 3) * 3;
					} else
						menaceb += (abs (f - FILE (kingsq[WHITE])) < 3);
				} else {
					++inactiveb;
					if ((castle & 3) == 0 && r < 3) {
						inactiveb += 5 * (FILE (kingsq[BLACK]) > 3 && f > FILE (kingsq[BLACK]));
						inactiveb += 5 * (FILE (kingsq[BLACK]) < 5 && f < FILE (kingsq[BLACK]));
					}
				}
				menaceb += 8 * (r > 5 && ((homer & 2) == 2 || RANK (kingsq[WHITE]) > 5));
				break;
			case BQ:
				f = FILE (sq);
				r = RANK (sq);
				if (lap[BLACK][f + 1] > 5) {	/* lap: black default=7
								 * white default=0  */
					if (r >= lap[WHITE][f + 1]) {
						menaceb += (abs (f - FILE (kingsq[WHITE])) < 3);
					}
					/*
					 * else menaceb += (abs(f -
					 * FILE(kingsq[WHITE])) < 3);
					 */
				} else if (r >= map[BLACK][f + 1]) {	/* map: black default=0
									 * white default=7  */
					if (r >= lap[WHITE][f + 1]) {
						menaceb += (abs (f - FILE (kingsq[WHITE])) < 3);
					}
					/*
					 * else menaceb += (abs(f -
					 * FILE(kingsq[WHITE])) < 3);
					 */
				}
				inactiveb += (r < 1);
				menaceb += 4 * (r > 5 && ((homer & 2) == 2 || RANK (kingsq[WHITE]) > 4));
				menaceb += 4 * (distance[128 + sq - kingsq[WHITE]] < 3);
				break;
			default:
				break;
			}

		}
	}
	/* WHITE */
	for (i = 0; i < 16; ++i) {
		sq = pieces[WHITE][i];
		if (sq) {
			p = board[sq];

			switch (p) {
			case WP:
				f = FILE (sq);
				r = RANK (sq);
				t = (r != map[WHITE][f + 1]);	/* doubled */
				
					weakw += 8 * t;
				
					if (t == 0 && r <= lap[BLACK][f + 1 - 1] && r <= lap[BLACK][f + 1 + 1] && r < lap[BLACK][f + 1]) {
					passerw += 10 * (passerw > 0);	/* multiple passer bonus */
					passerw += passed_pawn[7 - r];	/* base value */
					passerw += connected_passer[7 - r] * (board[sq - 1] == WP || board[sq + 1] == WP || board[sq + 15] == WP || board[sq + 17] == WP);
					passerw -= blocked_passer[7 - r] * (IS_BLACK (board[sq - 16]));	/* malis: blocked */
					passerw += (15 + 4 * (7 - r)) * (lar[WHITE][f + 1] > r);	/* bonus: rook behind */
					passerw += escorted_passer[7 - r] * (distance[128 + sq - 16 - kingsq[WHITE]] == 1);	/* bonus: king escort */
					passerw -= escorted_passer[7 - r] * (distance[128 + sq - 16 - kingsq[BLACK]] == 1);	/* malis: nme king */
				}
				weakw += 15 * (r > lap[WHITE][f + 1 - 1] && r > lap[WHITE][f + 1 + 1]);
				inactivew += (f > 1 && f < 6 && r == 6 && board[sq - 16] != MT);
				menacew += 4 * (f > 1 && f < 6 && r < 5);
				/* spacew += 3 * (7-r); */
				obstaclesw += bishopw[sq_color[sq]] * middle[sq];
				offensew += kingpawn[distance[128 + sq - 16 - kingsq[WHITE]]];
				offensew -= kingpawn[distance[128 + sq - 16 - kingsq[BLACK]]];
				break;
			case WN:
				f = FILE (sq);
				r = RANK (sq);
				if (((f > 1 && f < 6 && r > 0 && r < 5) || distance[128 + sq - kingsq[BLACK]] < 4) && lap[BLACK][f + 1 - 1] >= r && lap[BLACK][f + 1 + 1] >= r) {
					menacew += 2;
					if (board[sq + 17] == WP || board[sq + 15] == WP) {
						menacew += 3;
					}
				}
				inactivew += (f < 1) + (f > 6) + (r < 1) + 2 * (r > 6);
				break;
			case WB:
				f = FILE (sq);
				r = RANK (sq);
				if (((f > 1 && f < 6 && r > 0 && r < 5) || distance[128 + sq - kingsq[BLACK]] < 4) && lap[BLACK][f + 1 - 1] >= r && lap[BLACK][f + 1 + 1] >= r) {
					menacew += 2;
					if (board[sq + 17] == WP || board[sq + 15] == WP) {
						menacew += 3;
					}
				}
				inactivew += (f < 1) + (f > 6) + (r < 1) + 2 * (r > 6);
				break;
			case WR:
				f = FILE (sq);
				r = RANK (sq);
				if (lap[WHITE][f + 1] < 2) {	/* lap: black default=7
								 * white default=0 */
					if (r <= lap[BLACK][f + 1]) {
						rooksw += 15;
						menacew += (abs (f - FILE (kingsq[BLACK])) < 3) * 3;
					} else
						menacew += (abs (f - FILE (kingsq[BLACK])) < 3);
				} else if (r <= map[WHITE][f + 1]) {	/* map: black default=0
									 * white default=7 */
					if (r <= lap[BLACK][f + 1]) {
						/* rooksw += 10; */
						menacew += (abs (f - FILE (kingsq[BLACK])) < 3) * 3;
					} else
						menacew += (abs (f - FILE (kingsq[BLACK])) < 3);
				} else {
					++inactivew;
					if ((castle & 12) == 0 && r > 4) {
						inactivew += 5 * (FILE (kingsq[WHITE]) > 3 && f > FILE (kingsq[WHITE]));
						inactivew += 5 * (FILE (kingsq[WHITE]) < 5 && f < FILE (kingsq[WHITE]));
					}
				}
				menacew += 8 * (r < 2 && ((homer & 1) == 1 || RANK (kingsq[BLACK]) < 2));
				break;
			case WQ:
				f = FILE (sq);
				r = RANK (sq);
				if (lap[WHITE][f + 1] < 2) {	/* lap: black default=7
								 * white default=0 */
					if (r <= lap[BLACK][f + 1]) {
						menacew += (abs (f - FILE (kingsq[BLACK])) < 3);
					}
					/*
					 * else menacew += (abs(f -
					 * FILE(kingsq[BLACK])) < 3);
					 */
				} else if (r <= map[WHITE][f + 1]) {	/* map: black default=0
									 * white default=7 */
					if (r <= lap[BLACK][f + 1]) {
						menacew += (abs (f - FILE (kingsq[BLACK])) < 3);
					}
					/*
					 * else menacew += (abs(f -
					 * FILE(kingsq[BLACK])) < 3);
					 */
				}
				inactivew += (r > 6);
				menacew += 4 * (r < 2 && ((homer & 1) == 1 || RANK (kingsq[BLACK]) < 3));
				menacew += 4 * (distance[128 + sq - kingsq[BLACK]] < 3);
				break;
			default:
				break;
			}
		}
	}


	f = FILE (kingsq[BLACK]);
	r = RANK (kingsq[BLACK]);

	inactiveb += (f == 0 || f == 7 || r < 2);

	offenseb += pst[BK][kingsq[BLACK]];

	f = f + (f == 0) - (f == 7) + (f == 5) - (f == 2);
	safetyb += shield[lap[BLACK][f + 1 - 1]];	/* no pawn == 7 */
	safetyb += shield[lap[BLACK][f + 1 - 0]];
	safetyb += shield[lap[BLACK][f + 1 + 1]];

	safetyb -= 24 * (safetyb == 96);
	safetyb -= 16 * (safetyb < 31);

	safetyb -= storm[7 - map[WHITE][f + 1 - 1]];	/* map[W] = 7 for no
							 * pawn and rank1-6 */
	safetyb -= storm[7 - map[WHITE][f + 1]];
	safetyb -= storm[7 - map[WHITE][f + 1 + 1]];

	safetyb -= bk_defect[kingsq[BLACK]];

	f = FILE (kingsq[WHITE]);
	r = RANK (kingsq[WHITE]);

	inactivew += (f == 0 || f == 7 || r > 5);

	offensew += pst[WK][kingsq[WHITE]];

	f = f + (f == 0) - (f == 7) + (f == 5) - (f == 2);
	safetyw += shield[7 - lap[WHITE][f + 1 - 1]];	/* no pawn == 0 */
	safetyw += shield[7 - lap[WHITE][f + 1 - 0]];
	safetyw += shield[7 - lap[WHITE][f + 1 + 1]];

	safetyw -= 24 * (safetyw == 96);
	safetyw -= 16 * (safetyw < 31);

	safetyw -= storm[map[BLACK][f + 1 - 1]];	/* map[B] = 0 for no
							 * pawn and rank1-6 */
	safetyw -= storm[map[BLACK][f + 1]];
	safetyw -= storm[map[BLACK][f + 1 + 1]];

	safetyw -= wk_defect[kingsq[WHITE]];


	s += (safetyw * f_curve[force[BLACK]] / MF - safetyb * f_curve[force[WHITE]] / MF);
	s += (offensew * (MF - f_curve[force[BLACK]]) / MF - offenseb * (MF - f_curve[force[WHITE]]) / MF);


	s += (weakb * p_curve[force[WHITE]] / (MF / 2) - weakw * p_curve[force[BLACK]] / (MF / 2));

	s += (passerw * p_curve[force[BLACK]] / (MF / 2) - passerb * p_curve[force[WHITE]] / (MF / 2));

	s += (rooksw - rooksb);

	s += (inactive[inactiveb] - inactive[inactivew]);

	s += 3 * (obstaclesb - obstaclesw);

	s += (menace[menacew] - menace[menaceb]);

	/* s += (spacew - spaceb); */

	/*
	 * 584 Thur June 26 2008 f=material(&t); if(abs(f) < 100 &&
	 * (force[WHITE] + force[BLACK]) > 24) { f =
	 * (s-f)/(force[WHITE]+force[BLACK]);
	 * 
	 * if(f > best) { best = f; printf("***>>> %d %d<<<***\n", best, best *
	 * (force[WHITE]+force[BLACK])); } }
	 */

	estore (s);

	return ((stm) ? s + 5 : -s + 5);
}
