/***************************************************************************
*            init.c
*
*  Tue Apr  5 04:17:39 2005
*  Copyright  2005  DBW
*  opraus@yahoo.com
****************************************************************************/
  
#include "init.h"

const int quad[8] = { 0, 1, 2, 5, 12, 24, 32, 32 };



/*
	Passsed pawn is worth a Queen if unimpeded. But a pawn is rarely so, with 
	material on the board, or king within its square.

	King cannot stop more than one outside passer
	
	K, Q, and R can control both sq and advance.
	This is bad for the rook, as it promotes R in front of PP.
	Must be countered with penalty in Rook evaluation!

	K control is counted twice, .'. a PP's worst enemy or best friend!

	Pawn protection is counted twice .'. a PP best friend

	[][][][][][][][]
	[][][][][][][][]
	[][][][]Pb[][][]
	[][][]P3[][][][]  1 pp only. no connection
	[][]PwP2[][][][]  pp advance is pawn protected
	[][][]P1[][][][]  pp is pawn protected
	[][][][][][][][]
	[][][][][][][][]
*/ 
const int range[10][2] = { 
{4, 80},

{1, 20},
{1, 20},
{20, 20},

{1, 20},
{1, 15},

  /* Pawn and King counted twice! */ 
  /* King fulfills both rolls! */ 

{4, 30},
{4, 30},

{0, 20},
{0, 20}
};


int
scale (int min, int max, int i) 
{
  
return min + quad[i] * (max - min) / quad[6];
}




#define ISOBASE 4
  
/*
	outside values are dimished because weaknesses are likely to be
	unrelated to rest of score. Pawn structure tends to coordinate
	SPACE and other pawn considerations - eg. OUTPOSTS, Bishop mobility.
	So, 'Push pawns!' without 'Keep ranks!' wouldnt work
	
	Pushing rook pawns offers little space
	Rook pawns are less likely under overwelming attack
	Rook pawns attack 1 relatively unimportant file [excepting shield or storm]

	If Rook pawns are 15% less valuable, then their weaknesses ought to be also

	Center pawns are 20-30% more valuable than 
*/ 

int iso_by[2][8] = { 
{0, 3, 4, 4, 4, 5, 5, 0}, 
{1, 2, 2, 2, 2, 2, 2, 1} 
};


/*
	pieces on own back ranks can have no posotive value.
	they are merely defending
*/ 
int lines[8][2][8] = { 

{{0, 0, 0, 0, 0, 0, 0, 0}, 
{0, 0, 0, 0, 0, 0, 0, 0}}, 

{{0, 0, 4, 8, 12, 16, 20, 0},
{0, 0, 0, 0, 0, 0, 0, 0}}, 

{{-11, 0, 0, 0, 0, 0, 0, 0}, 
{0, 0, 0, 0, 0, 0, 0, 0}}, 

{{-20, -5, 0, 5, 10, 10, 10, 0}, 
{-10, 0, 5, 5, 5, 5, 0, -10}}, 

{{-20, -5, 0, 5, 10, 10, 10, 0}, 
{-10, 0, 5, 5, 5, 5, 0, -10}}, 

{{-8, 0, 0, 0, 0, 0, 0, 0},
{0, 0, 0, 0, 0, 0, 0, 0}},

{{1, 2, 0, 0, 0, 0, 3, 3},
{-1, -1, 5, 6, 6, 5, -1, -1}},

{{0, 0, 0, 0, 0, 0, 0, 0}, 
{0, 0, 0, 0, 0, 0, 0, 0}} 
};


/*
PST serves 2 main purposes:
1. bad placement/development/activity/mobility*
piece [even Q] on edge, is less likely to be able to get to other side
do not chose metrics at expense of Q out of game e.g.
Must be careful to balance PST values, but they are not 0
2. move ordering   
abs PST for pawn must be high
delta PST for knights must be greatest

Notes:
sliders can 'improve' their position 'faster' than steppers or hoppers
however, it is usually more important to improve non-sliders first
so, eg., 1 'hop' of knight off back should be > bishop sliding to center

pst[t][A1+r*16+f] = s;
pst[t+1][H8-r*16-f] = s;
*/ 
  
void
init_iso ()
{
  
int r, f;
  
for (r = 0; r < 8; ++r)
    {
      
#ifdef SHOWTABLES        
	printf ("\n");
      
#endif	/* 
 */
	for (f = 0; f < 8; ++f)
	{
	  
isolated_pst[0][A1 + 16 * r + f] =
	    ISOBASE + iso_by[1][f] + iso_by[0][r];
	  
isolated_pst[1][A8 - 16 * r + f] =
	    ISOBASE + iso_by[1][f] + iso_by[0][r];
	  
#ifdef SHOWTABLES  
	    printf ("%d\t", ISOBASE + iso_by[1][f] + iso_by[0][r]);
	  
#endif	/* 
 */
	}
    
}
  
#ifdef SHOWTABLES  
    printf ("\n\n");
  
#endif	/* 
 */

}



void
init_pst ()
{
  
int r, f, p, s, sq;
  

for (p = 1; p < 8; ++p)
    {
      for (r = 0; r < 8; ++r)
	{
	  
for (f = 0; f < 8; ++f)
	    {
	      
sq = A1 + (r << 4) + f;
	      

if (p == 1)
		
		{
		  

s = (1 + (f > 0 && f < 7)) * (r - 1);
		  

s += 6 * (f > 2 && f < 5 && r > 1 && r < 5);
		  

s -= 2 * (sq == D3);
		  

s += 2 * (f > 1 && f < 6 && r > 1 /* && r < 5 */ );
		  

s += 1 * (sq == C4);
		  
s += 1 * (sq == C5);
		  
s += 1 * (sq == C3);
		  
s += 1 * (sq == F3);
		  

s += 4 * (f > 1 && f < 6 && r > 4);
		
}
	      
	      else if (p == 2)
		{
		  
s = lines[p][0][r] + lines[p][1][f];
		  
s += outposts[0][sq] / 3;
		  
		    /*
		       00000000
		       0xxxxxx0
		       0x1111x0
		       0x1111x0
		       0x1111x0
		       0x1111x0
		       0xxxxxx0
		       00000000
		     */ 
		    s += 8 * (f > 1 && f < 6 && r > 1 && r < 7);
		  

s += 5 * (f > 0 && f < 7 && r > 1 && r < 7);
		  

s -= 4 * corner[0][sq];
		  
s -= 2 * corner[1][sq];
		  
s += 2 * (sq == C3);
		  
s += 2 * (sq == F3);
		  
s += 1 * (sq == E2);
		  
s += 1 * (sq == D2);
		
}
	      
	      else if (p == 5)
		{
		  
s = lines[p][0][r] + lines[p][1][f] - 1;
		  
s += outposts[0][sq] / 12;
		  
s += 1 * (f > 1 && f < 6 && r > 2 && r < 7);
		  

s += 1 * (f > 0 && f < 7 && r > 1 && r < 7);
		  

s +=
		    4 * (min (abs (f - r + 1), abs ((7 - f) - r + 1)) < 3);
		  
s += 1 * (sq == G2);
		

}
	      
	      else if (p == 7)
		{
		  s =
		    (pst[10][A1 + r * 16 + f] + pst[12][A1 + r * 16 + f]) / 2;
		  s += 2 * (sq == C2);
		  
s += 2 * (sq == D2);
		  
s += 2 * (sq == E2);
		  
s -= 4 * (r == 0);
		  
s -= 4 * (sq == A1);
		  
s -= 4 * (sq == B1);
		  
s -= 4 * (sq == G1);
		  
s -= 4 * (sq == H1);
		
}
	      
	      else
		{
		  
s = lines[p][0][r] + lines[p][1][f];
		
}
	      
#ifdef SHOWTABLES                
		printf ("%d\t", s);
	      
#endif	/* 
 */
		pst[p * 2][A1 + r * 16 + f] = s;
	      
pst[p * 2 + 1][A8 - r * 16 + f] = s;
	    
}
	  
#ifdef SHOWTABLES              
	    printf ("\n");
	  
#endif	/* 
 */
	}
      
#ifdef SHOWTABLES          
	printf ("%d\n\n", p);
      
#endif	/* 
 */
    }
  

for (r = 0; r < 8; ++r)
    {
      
for (f = 0; f < 8; ++f)
	{
	  
s = lines[4][0][r] + lines[4][1][f];
	  
#ifdef SHOWTABLES              
	    printf ("%d\t", s);
	  
#endif	/* 
 */
	    k_to_war[0][A1 + r * 16 + f] = s;
	  
k_to_war[1][A8 - r * 16 + f] = s;
	
}
      
#ifdef SHOWTABLES          
	printf ("\n");
      
#endif	/* 
 */
    }
  
#ifdef SHOWTABLES      
    printf ("\n\n");
  
#endif	/* 
 */
}




int
pseudo (int f, int t)
{
  
int dx = (f & 7) - (t & 7);
  
int dy = (f >> 4) - (t >> 4);
  
if (dx == 0 || dy == 0)
    return 1;
  
if (dx < 0)
    dx = -dx;
  
if (dy < 0)
    dy = -dy;
  
if (dx == dy)
    return 1;
  
if ((dx + dy) == 3)
    return 1;
  
return 0;

}


int
vector (int f, int t)
{
  
int dy = (t - f) / 16;
  
int dx = (t - f) - (dy * 16);
  
if (dx == 0)
    return (dy > 0) ? 16 : -16;
  
if (dy == 0)
    return (dx > 0) ? 1 : -1;
  

if (dx == dy)
    return (dx > 0) ? 17 : -17;
  
if (abs (dx) == abs (dy))
    return (dx > 0) ? -15 : 15;
  
return 0;

}



void
init_delta_table ()
{
  
int d, x, y;
  
for (x = -7; x < 8; x++)
    {
      
for (y = -7; y < 9; y++)
	{
	  
d = x * 16 + y;
	  
distance[128 + d] = max (abs (x), abs (y));
	  
pseudoev[128 + d] = pseudo (0, 16 * abs (x) + abs (y));
	  
taxicab[128 + d] = abs (x) + abs (y);
	  
manhattan[128 + d] = min (abs (x), abs (y));
	  
diagonalm[128 + d] = abs (abs (x) - abs (y));
	
}
    
}
  

    /*
       printf("\n");
       for(x = -7; x < 8; x++){
       for(y = -7; y < 9; y++){
       d = x*16+y;


       printf("%d ", distance[128+d]);




       }
       printf("\n");
       }
       
       printf("\n");
       for(x = 0; x < 256; ++x)
       {
       printf("%1d ", distance[x]);
       if((x&15)==15) printf("\n");
       
       }
       printf("\n");
     */ 
}




/*
void init_p_curve(void)
{
	int i;
	for(i=0;i<64;++i)
	{
		p_curve[i] = MF-f_curve[i];

	}
}
*/ 
